use crate::cli::prelude::*;

mod config;
mod gitlab;
mod jira;
mod prelude;

// TODO: clap for CLI

#[derive(Parser)]
#[command(version, about, long_about = None)]
#[command(propagate_version = true)]
pub struct Cli {
    #[command(subcommand)]
    pub command: Commands,

    /// Define which tracing values are logged
    /// See https://docs.rs/tracing-subscriber/latest/tracing_subscriber/filter/struct.EnvFilter.html
    /// for details
    #[arg(long, default_value = None)]
    pub tracing_env_filter: Option<String>,

    /// The directory to write log files (defaults to $XDG_CACHE_HOME/taskr/logs)
    #[arg(long, default_value = None)]
    pub logs_directory: Option<String>,

    /// The directory that data is stored in (defaults to $XDG_DATA_HOME/taskr/data)
    #[arg(long, default_value = None)]
    pub data_directory: Option<String>,

    /// The configuration file to load (defaults to $XDG_CONFIG_HOME/taskr/config.toml)
    #[arg(long, default_value = None)]
    pub config_file_path: Option<String>,
}

impl Cli {
    pub async fn exec(self) -> Result<()> {
        let tasks = Arc::new(crate::tasks::Tasks::try_new(
            self.config_file_path,
            self.data_directory,
        )?);

        match self.command {
            Commands::Ui(args) => args.run(tasks).await,
            Commands::Purge(args) => args.run(tasks).await,
            Commands::Cleanup(args) => args.run(tasks).await,
            Commands::Sync(args) => args.sync(tasks).await,
            Commands::List(args) => args.list(tasks).await,
            Commands::Jira(jira) => jira.exec(tasks).await,
            Commands::Gitlab(gitlab) => gitlab.exec(tasks).await,
            Commands::Config(config) => config.exec(tasks).await,
        }
    }
}

#[derive(Subcommand)]
pub enum Commands {
    /// List local tasks with options for syncing
    #[clap(visible_alias = "ls")]
    List(list::Args),

    /// Sync local tasks with GitLab and Jira
    Sync(sync::Args),

    /// Purge all local data
    Purge(PurgeArgs),

    /// Performs various cleanup operations on local data
    Cleanup(CleanupArgs),

    /// Run the interactive terminal user interface (TUI)
    #[clap(visible_alias = "tui")]
    Ui(ui::Args),

    /// Interact with the configured Jira endpoint
    #[command(subcommand)]
    Jira(jira::Command),

    /// Interact with the configured GitLab endpoint
    #[command(subcommand)]
    Gitlab(gitlab::Command),

    /// Interact with the configured GitLab endpoint
    #[command(subcommand)]
    Config(config::Command),
}

#[derive(Args)]
pub struct PurgeArgs {}
impl PurgeArgs {
    pub async fn run(&self, tasks: SharedTasks) -> Result<()> {
        tasks.purge_all()?;
        println!("Local data purged!");
        Ok(())
    }
}

#[derive(Args)]
pub struct CleanupArgs {}
impl CleanupArgs {
    pub async fn run(&self, tasks: SharedTasks) -> Result<()> {
        tasks.cleanup_all()?;
        println!("Local data cleaned up!");
        Ok(())
    }
}

mod ui {
    use crate::cli::prelude::*;

    #[derive(Parser, Debug)]
    pub struct Args {}

    impl Args {
        #[instrument]
        pub async fn run(&self, tasks: SharedTasks) -> Result<()> {
            crate::tui::Tui::run(tasks).await
        }
    }
}

mod list {
    use crate::{cli::prelude::*, task::Task};

    #[derive(Parser)]
    pub struct Args {
        /// Whether or not to sync before listing tasks
        #[arg(short, long, default_value = None)]
        pub sync: bool,
    }

    impl Args {
        pub async fn list(&self, tasks: SharedTasks) -> Result<()> {
            if self.sync {
                eprintln!("Syncing...");
                tasks.sync().await?;
            }
            // TODO: if we _don't_ sync, check last sync time and let user know
            // that things may have changed
            let mut tasks: Vec<Task> = tasks.all()?.into_values().collect();
            tasks.sort_unstable();
            eprintln!("{} Tasks", tasks.len());
            // TODO: make this generical? take a vec of vecs or something, scan each
            // entry and their lengths and we can spit out a nice table with each
            // "column" having equal length
            for t in tasks {
                println!("{t}");
            }
            Ok(())
        }
    }
}

mod sync {
    use crate::cli::prelude::*;

    #[derive(Parser)]
    pub struct Args {}

    impl Args {
        pub async fn sync(&self, tasks: SharedTasks) -> Result<()> {
            eprintln!("Syncing...");
            tasks.sync().await?;
            eprintln!("Done!");
            Ok(())
        }
    }
}

impl Cli {
    pub fn new() -> Self {
        Self::parse()
    }
}

#[test]
fn verify_cli() {
    use clap::CommandFactory;
    Cli::command().debug_assert()
}
