use cliclack::{input, password};
use futures::StreamExt;
use signal_hook::consts::*;
use signal_hook_tokio::Signals;

use crate::cli::prelude::*;

#[derive(Subcommand)]
pub enum Command {
    /// Write the current config (without secrets) in TOML format
    Show,

    /// Interactively setup a Taskr configuration
    Setup,
}

impl Command {
    pub async fn exec(&self, tasks: SharedTasks) -> Result<()> {
        match self {
            Command::Show => self.show(tasks).await,
            Command::Setup => Self::setup(tasks).await,
        }
    }

    pub async fn show(&self, tasks: SharedTasks) -> Result<()> {
        println!("{}", toml::to_string_pretty(tasks.config()?)?);
        Ok(())
    }

    pub async fn setup(stasks: SharedTasks) -> Result<()> {
        use cliclack::{intro, outro};

        let tasks = stasks.clone();
        let cliclack_handle: tokio::task::JoinHandle<Result<()>> = tokio::spawn(async move {
            // TODO: maybe use the normal builder
            // TODO: handle sigint and reset the terminal gracefully

            intro("Configure taskr")?;

            let gitlab_url: String = input("What is your GitLab URL?")
                .placeholder("https://gitlab.com")
                .validate_interactively(|input: &String| match url::Url::parse(input) {
                    Ok(_) => Ok(()),
                    Err(_) => Err("Must be a URL"),
                })
                .interact()?;

            let jira_url: String = input("What is your Jira URL?")
                .placeholder("https://company.atlassian.net")
                .validate_interactively(|input: &String| match url::Url::parse(input) {
                    Ok(_) => Ok(()),
                    Err(_) => Err("Must be a URL"),
                })
                .interact()?;

            let mut config = Config::build(
                Config::builder(&tasks.config_file_path)?
                    .set_override("gitlab.url", gitlab_url.as_str())?
                    .set_override("jira.url", jira_url.as_str())?,
            )?;

            // TODO: link user to page to setup tokens
            config.secrets.gitlab_token =
                password("What is your GitLab Personal Access Token? (scopes: read_api)")
                    // .placeholder("glpat-XXXX...")
                    .interact()?;

            config.secrets.jira_token = password("What is your Jira API Token?").interact()?;

            config.save(&tasks.config_file_path)?;

            // TODO: initialize known keyring values/entries
            outro("All set! Try `taskr list --sync` now!")?;
            Ok(())
        });

        let mut signals = Signals::new(&[SIGHUP, SIGTERM, SIGINT, SIGQUIT])?;
        let signal_handle = signals.handle();

        // TODO: either we will die, get a signal, or we setup

        tokio::select! {
            _ = signals.next() => {
                // outro_cancel("Canceled taskr setup")?;
            }
            _ = cliclack_handle => {
                signal_handle.close();
            }
        };

        // reset our terminal?

        Ok(())
    }
}
