use crate::result::Result;
use reqwest::{Client as RClient, Method, Url};
use reqwest_middleware::{ClientBuilder, ClientWithMiddleware, RequestBuilder};
use reqwest_retry::{policies::ExponentialBackoff, RetryTransientMiddleware};
use reqwest_tracing::TracingMiddleware;
use serde::de;

#[derive(Debug)]
pub struct Client {
    client: ClientWithMiddleware,
    base_url: Url,
}

pub trait ResourceRequest {
    async fn res<T: de::DeserializeOwned>(self) -> Result<T>;
}

impl ResourceRequest for RequestBuilder {
    async fn res<T: de::DeserializeOwned>(self) -> Result<T> {
        // TODO: make this a field on this struct or something?
        /*
        use tracing::debug;
        let body = self.send().await?.text().await?;
        debug!(body);
        serde_json::from_str(&body).map_err(|e| e.into())
        */

        self.send().await?.json().await.map_err(|e| e.into())
    }
}

impl Client {
    pub fn try_new(client: RClient, base_url: &str) -> Result<Self> {
        let retry_policy = ExponentialBackoff::builder().build_with_max_retries(3);

        let client = ClientBuilder::new(client)
            .with(TracingMiddleware::default())
            .with(RetryTransientMiddleware::new_with_policy(retry_policy))
            .build();

        // force trailing slash
        let mut base_url = Url::parse(base_url)?;
        base_url.set_path(&format!("{}{}", base_url.path().trim_end_matches('/'), "/"));

        Ok(Self { client, base_url })
    }

    /// Takes a path segment and properly joins it relative to `base_url`
    pub fn url(&self, path: &str) -> Result<Url> {
        self.base_url
            .join(path.trim_start_matches('/'))
            .map_err(|e| e.into())
    }

    pub fn build(&self, method: Method, rel_url_path: &str) -> Result<RequestBuilder> {
        Ok(self.client.request(method, self.url(rel_url_path)?))
    }

    pub async fn request<T: de::DeserializeOwned>(
        &self,
        method: Method,
        rel_url_path: &str,
    ) -> Result<T> {
        self.build(method, rel_url_path)?.res().await
    }

    pub async fn get<T: de::DeserializeOwned>(&self, rel_url_path: &str) -> Result<T> {
        self.request::<T>(Method::GET, rel_url_path).await
    }
}
