use crate::prelude::*;

use std::{collections::HashMap, fmt::Debug, fs::File, io::Write, path::Path};

use color_eyre::{eyre::Context, Section};
use config::{builder::DefaultState, Config as CConfig, ConfigBuilder, Value};
use serde::{Deserialize, Serialize};

const CURRENT_VERSION: u64 = 1;

#[derive(Serialize, Deserialize, Debug, Clone)]
pub struct Gitlab {
    pub url: String,
}

#[derive(Serialize, Deserialize, Debug, Clone)]
pub struct Jira {
    pub url: String,
}

#[derive(Deserialize, Debug, Clone, Default)]
pub struct Secrets {
    pub jira_token: String,
    pub gitlab_token: String,
}

#[derive(Serialize, Deserialize, Debug, Clone)]
pub struct Config {
    pub version: u64,
    pub gitlab: Gitlab,
    pub jira: Jira,

    #[serde(skip_serializing)]
    pub secrets: Secrets,
}

impl Config {
    pub fn load(config_file_path: &Path) -> Result<Self> {
        let mut conf = Self::build(Self::builder(config_file_path)?)?;

        if conf.secrets.jira_token == "" {
            conf.secrets.jira_token = keyring::Entry::new("taskr", "jira_token")?.get_password()?;
        }

        if conf.secrets.gitlab_token == "" {
            conf.secrets.gitlab_token =
                keyring::Entry::new("taskr", "gitlab_token")?.get_password()?;
        }

        Ok(conf)
    }

    /// This should inject the files/values/whatever into their appropriate
    /// places such that the next time Self::load() is called (with the same
    /// `config_file_path`) you end up with the same Self.
    pub fn save(&self, config_file_path: &Path) -> Result<()> {
        if let Some(p) = config_file_path.parent() {
            std::fs::create_dir_all(p)?;
        }
        let mut file = File::create(config_file_path)?;
        file.write_all(toml::to_string(self)?.as_bytes())?;

        keyring::Entry::new("taskr", "gitlab_token")?.set_password(&self.secrets.gitlab_token)?;
        keyring::Entry::new("taskr", "jira_token")?.set_password(&self.secrets.jira_token)?;

        Ok(())
    }

    pub fn build(builder: ConfigBuilder<DefaultState>) -> Result<Self> {
        builder
            .build()?
            .try_deserialize()
            .suggestion("run `taskr config setup` to ensure valid configuration")
            .with_context(|| "failed to deserialize configuration")
    }

    pub fn builder(config_file_path: &Path) -> Result<ConfigBuilder<DefaultState>>
where {
        Ok(Self::default_builder()?
            .add_source(config::File::from((*config_file_path).as_ref()).required(false))
            .add_source(config::Environment::with_prefix("taskr").separator("__")))
    }

    pub fn default_builder() -> Result<ConfigBuilder<DefaultState>> {
        Ok(CConfig::builder()
            .set_default("version", CURRENT_VERSION)?
            .set_default("secrets", HashMap::<String, Value>::new())?
            .set_default("secrets.gitlab_token", "")?
            .set_default("secrets.jira_token", "")?
            .set_default("jira", HashMap::<String, Value>::new())?
            .set_default("gitlab", HashMap::<String, Value>::new())?
            .set_default("jira.url", "")?
            .set_default("gitlab.url", "")?)
    }

    #[allow(dead_code)]
    pub fn defaults() -> Result<Self> {
        Self::build(Self::default_builder()?)
    }
}

#[cfg(test)]
mod tests {
    use crate::config::*;

    #[test]
    fn default_config_toml_deserialize_to_valid_config() -> Result<()> {
        let tmp = std::env::temp_dir()
            .join("taskr_defaults.default_config_toml_deserialize_to_valid_config.toml");

        std::fs::write(
            &tmp,
            r#"[gitlab]
            url = "https://gitlab.com"
            
            [jira]
            url = "https://jira.com"

            [secrets]
            gitlab_token = "asdf"
            jira_token = "1337"
            "#,
        )?;

        println!("Defaults: {:?}", Config::load(&tmp)?);
        let _ = std::fs::remove_file(tmp);
        Ok(())
    }
}
