use crate::prelude::*;

use std::path::Path;
use tracing_appender::non_blocking::WorkerGuard;
use tracing_subscriber::{filter::LevelFilter, EnvFilter};

pub fn setup_error_handling() -> Result<()> {
    color_eyre::install()?;
    Ok(())
}

pub fn setup_logging<T, S>(
    logs_directory: &Option<T>,
    env_filter_directive: &Option<S>,
) -> Result<WorkerGuard>
where
    T: AsRef<Path>,
    S: AsRef<str>,
{
    let mut default_logs_dir = Default::default();
    let logs_dir = logs_directory
        .as_ref()
        .map(T::as_ref)
        .map(Result::Ok)
        .unwrap_or_else(|| {
            // really want to avoid calling this so it's put in here to force it to be lazy
            // also have to shove it in a function-scoped variable so the
            // reference we create in this closure has the same lifetime as the function
            default_logs_dir = xdg::BaseDirectories::new()?.create_cache_directory("taskr/logs")?;
            Ok(default_logs_dir.as_ref())
        })?;

    eprintln!("Logs Directory: {}", logs_dir.display());
    if !logs_dir.exists() {
        std::fs::create_dir_all(&logs_dir)?;
    }

    let file_appender = tracing_appender::rolling::daily(logs_dir, "log");
    let (non_blocking, guard) = tracing_appender::non_blocking(file_appender);

    let env_filter = env_filter_directive
        .as_ref()
        .map(S::as_ref)
        .unwrap_or("taskr=trace,info");

    let filter = EnvFilter::builder()
        .with_default_directive(LevelFilter::TRACE.into())
        .parse_lossy(env_filter);

    eprintln!("EnvFilter Directive: {}", filter);

    tracing_subscriber::fmt()
        .with_writer(non_blocking)
        .with_env_filter(filter)
        .init();

    Ok(guard)
}
