use crate::prelude::*;

use std::{
    collections::{HashMap, HashSet},
    env,
    hash::RandomState,
    path::{Path, PathBuf},
    process::Command,
    sync::OnceLock,
};
use tracing::{info, warn};

use crate::task::Task;

use sled::IVec;

use crate::{gitlab::GitLab, jira::Jira, result::Result};

pub type Db = sled::Db;

#[derive(Debug)]
pub struct Tasks {
    pub data_dir: PathBuf,
    pub config_file_path: PathBuf,

    config: OnceLock<Config>,
    gitlab: OnceLock<GitLab>,
    jira: OnceLock<Jira>,
    db: OnceLock<Db>,
}

const MY_DONE_STATUS_CATEGORY_KEY: &str = "done";

impl Tasks {
    pub fn try_new<D>(config_file_path: Option<D>, data_dir: Option<D>) -> Result<Self>
    where
        D: AsRef<Path>,
    {
        debug!("Creating Tasks instance...");
        let db = OnceLock::new();
        let config = OnceLock::new();
        let gitlab = OnceLock::new();
        let jira = OnceLock::new();

        let xdgonce = OnceLock::new();
        let xdg = || -> Result<&xdg::BaseDirectories> {
            match xdgonce.get() {
                Some(x) => Ok(x),
                None => {
                    let result = xdg::BaseDirectories::new()?;
                    let _ = xdgonce.set(result);
                    Ok(xdgonce.get().unwrap())
                }
            }
        };

        let config_file_path: PathBuf = if let Some(p) = config_file_path {
            p.as_ref().to_owned()
        } else {
            xdg()?.get_config_file("taskr/config.toml")
        };

        let data_dir: PathBuf = if let Some(p) = data_dir {
            p.as_ref().to_owned()
        } else {
            xdg()?.create_data_directory("taskr/data")?
        };

        Ok(Self {
            config_file_path,
            config,
            gitlab,
            jira,
            data_dir,
            db,
        })
    }

    pub fn config(&self) -> Result<&Config> {
        match self.config.get() {
            Some(c) => Ok(c),
            None => {
                let result = Config::load(self.config_file_path.as_path())?;
                let _ = self.config.set(result);
                Ok(self.config.get().unwrap())
            }
        }
    }

    pub fn db(&self) -> Result<&Db> {
        match self.db.get() {
            Some(d) => Ok(d),
            None => {
                let result = sled::open(&self.data_dir)?;
                let _ = self.db.set(result);
                Ok(self.db.get().unwrap())
            }
        }
    }

    pub fn gitlab(&self) -> Result<&GitLab> {
        match self.gitlab.get() {
            Some(g) => Ok(g),
            None => {
                // TODO: find a way to more-lazily load the token?
                let gitlab_token_entry = keyring::Entry::new("taskr", "gitlab_token")?;

                let gl_token = match gitlab_token_entry.get_password() {
                    Ok(token) => {
                        info!("GitLab token loaded from keyring");
                        token
                    }
                    Err(_) => {
                        let token = match env::var("GITLAB_TOKEN") {
                            Ok(token) => {
                                info!("GitLab token loaded from environment variable GITLAB_TOKEN");
                                token
                            }
                            Err(_) => {
                                let output = Command::new("pass")
                                    .arg("client/divvy/gitlab-glpat")
                                    .output()?;

                                let result =
                                    String::from_utf8_lossy(&output.stdout).trim().to_string();
                                info!("GitLab token loaded from password-store entry client/divvy/gitlab-glpat");
                                result
                            }
                        };
                        info!("GitLab token stored in keyring");
                        gitlab_token_entry.set_password(&token)?;
                        token
                    }
                };
                let result =
                    GitLab::try_new(&format!("{}/api/v4", self.config()?.gitlab.url), &gl_token)?;
                let _ = self.gitlab.set(result);
                Ok(self.gitlab.get().unwrap())
                // TODO: ensure the token works?
            }
        }
    }

    pub fn jira(&self) -> Result<&Jira> {
        match self.jira.get() {
            Some(j) => Ok(j),
            None => {
                // TODO: find a way to more-lazily load the token?
                let jira_token_entry = keyring::Entry::new("taskr", "jira_token")?;

                let jira_token = match jira_token_entry.get_password() {
                    Ok(token) => {
                        info!("Jira token loaded from keyring");
                        token
                    }
                    Err(_) => {
                        let token = match env::var("JIRA_TOKEN") {
                            Ok(token) => {
                                info!("Jira token loaded from environment variable JIRA_TOKEN");
                                token
                            }
                            Err(_) => {
                                let output = Command::new("pass")
                                    .arg("client/divvy/jira-api-token-with-email")
                                    .output()?;

                                let result =
                                    String::from_utf8_lossy(&output.stdout).trim().to_string();
                                info!("Jira token loaded from password-store entry client/divvy/jira-api-token-with-email");
                                result
                            }
                        };
                        info!("GitLab token stored in keyring");
                        jira_token_entry.set_password(&token)?;
                        token
                    }
                };

                let result = Jira::try_new(&format!("{}", self.config()?.jira.url), &jira_token)?;
                let _ = self.jira.set(result);
                Ok(self.jira.get().unwrap())
                // TODO: ensure the token works?
            }
        }
    }

    pub fn all(&self) -> Result<HashMap<String, Task>> {
        let mut result = HashMap::new();
        let scan = self.db()?.open_tree("tasks")?.scan_prefix(&[]);
        for entry in scan {
            match entry {
                Ok((_key, val)) => {
                    let task = Task::try_from(val)?;
                    result.insert(task.jira_key.to_owned(), task);
                }
                Err(e) => return Err(e.into()),
            }
        }
        Ok(result)
    }

    #[allow(dead_code)]
    pub fn get(&self, key: &str) -> Result<Option<Task>> {
        let ivec = self.db()?.open_tree("tasks")?.get(key)?;
        match ivec {
            Some(v) => Ok(Some(Task::try_from(v)?)),
            None => Ok(None),
        }
    }

    pub fn save(&self, task: &Task) -> Result<()> {
        let ivec: IVec = task.try_into()?;
        let _previous_value = self
            .db()?
            .open_tree("tasks")?
            .insert(&task.jira_key, ivec)?;
        self.db()?.flush()?;
        Ok(())
    }

    pub fn delete(&self, task: &Task) -> Result<()> {
        let _previous_value = self.db()?.open_tree("tasks")?.remove(&task.jira_key)?;
        self.db()?.flush()?;
        Ok(())
    }

    pub fn purge_all(&self) -> Result<()> {
        warn!("Purging all local tasks...");
        for (_, t) in self.all()? {
            self.delete(&t)?;
        }
        self.db()?.flush()?;
        Ok(())
    }

    pub fn cleanup(&self, task: &mut Task) -> Result<()> {
        static RE: OnceLock<Regex> = OnceLock::new();
        let regex = RE.get_or_init(|| Regex::new(r"^\s*\:[a-zA-Z0-9_-]+\:\s*").unwrap());
        let mut changed = false;
        if let Some(m) = regex.find(&task.description) {
            task.description = task.description[m.range().end..].to_owned();
            changed = true;
        }
        if task.status.starts_with("Blocked") {
            task.status = "Blocked".to_owned();
            changed = true;
        }
        if task.status.starts_with("In Development") {
            task.status = "InProgress".to_owned();
            changed = true;
        }
        if changed {
            self.save(task)?;
        }
        Ok(())
    }

    pub fn cleanup_all(&self) -> Result<()> {
        for (_, mut t) in self.all()? {
            self.cleanup(&mut t)?;
        }
        self.db()?.flush()?;
        Ok(())
    }

    /// for a task that has no associated open and appropriately assigned jira issue
    async fn fix_dangling_task(&self, task: &Task) -> Result<()> {
        // check if closed, if it is, delete the task
        let issue = self.jira()?.issue(&task.jira_key).await?;
        if issue.fields.status.status_category.key == MY_DONE_STATUS_CATEGORY_KEY {
            info!(
                "Deleting task {} due to being marked {} in Jira",
                task.jira_key, MY_DONE_STATUS_CATEGORY_KEY
            );
            // remote issue is done, so delete the task
            self.delete(task)?;
        }
        Ok(())
    }

    /// fetches jira issues and compares to known local tasks
    /// for use when sync'ing local tasks to remote state (jira, gitlab)
    pub async fn sync(&self) -> Result<()> {
        let mut tasks = self.all()?;
        let issues = crate::jira::Jira::by_key(self.jira()?.assigned_open_issues().await?);
        let task_keys: HashSet<String, RandomState> =
            HashSet::from_iter(tasks.keys().map(|s| s.to_owned()));
        let issue_keys: HashSet<String, RandomState> =
            HashSet::from_iter(issues.keys().map(|s| s.to_owned()));

        // keys which have a task but not an issue
        let dangling_keys: HashSet<String, RandomState> = HashSet::from_iter(
            task_keys
                .difference(&issue_keys)
                .into_iter()
                .map(|s| s.to_owned()),
        );

        info!("Creating new tasks from issues without a task");
        for key in issue_keys.difference(&task_keys) {
            let issue = issues.get(key).unwrap();
            let mut task: Task = issue.try_into()?;
            self.cleanup(&mut task)?;
            self.save(&task)?;
            info!("Created task {}", task);
            tasks.insert(task.jira_key.clone(), task);
        }

        futures::future::try_join_all(
            dangling_keys
                .iter()
                .map(|t| self.fix_dangling_task(tasks.get(t).unwrap())),
        )
        .await?;

        // TODO: blocking? maybe should be async?
        // while let Ok(_res) = rx.recv() {
        // awaiting all the tasks in the joinset
        // }

        // fix_missing_task(

        // let mut futures = Vec::with_capacity((results.total / results.max_results) + 1);

        // for i in 1..futures.capacity() {
        //     let start_at = i * results.max_results;
        //     debug!("Fetching issues for jql {} (page {})", jql, i);
        //     futures.push(spawn(
        //         self.client
        //             .build(Method::GET, "/rest/api/3/search")?
        //             .query(&[("jql", jql), ("startAt", &start_at.to_string())])
        //             .res::<IssueSearch>(),
        //     ));
        // }

        info!("Checking Jira issues for updates for existing tasks");
        info!("Checking missing Jira issues for updates for existing tasks");

        // save last sync date
        let now = chrono::Local::now();

        self.db()?
            .open_tree("task:meta")?
            .insert(b"last_sync_date", serde_json::to_vec(&now)?)?;

        Ok(())
    }
}
