use crate::prelude::*;

use crossterm::{
    event::{self, KeyCode, KeyEventKind, KeyModifiers},
    terminal::{disable_raw_mode, enable_raw_mode, EnterAlternateScreen, LeaveAlternateScreen},
    ExecutableCommand,
};
use futures::stream::StreamExt;
use futures::FutureExt;
use ratatui::{
    prelude::{CrosstermBackend, Stylize, Terminal},
    widgets::Paragraph,
};
use signal_hook::consts::*;
use signal_hook_tokio::Signals;
use std::io::{stdout, Stdout};

pub struct Tui {}

enum Event {
    CrosstermEvent(crossterm::event::Event),
    // CrosstermError(Arc<dyn std::error::Error>),
    Quit,
    Update,
    Render,
}

impl Tui {
    #[instrument]
    fn setup_screen() -> Result<()> {
        debug!("Setting up stdout for TUI...");
        stdout().execute(EnterAlternateScreen)?;
        enable_raw_mode()?;
        Ok(())
    }

    fn new_terminal() -> Result<Terminal<CrosstermBackend<Stdout>>> {
        debug!("Initializing TUI Terminal with Crossterm backend...");
        let mut terminal = Terminal::new(CrosstermBackend::new(stdout()))?;
        terminal.clear()?;
        return Ok(terminal);
    }
    pub async fn run(tasks: SharedTasks) -> Result<()> {
        Self::setup_screen()?;
        let mut terminal = Self::new_terminal()?;
        let (tx, mut rx) = tokio::sync::mpsc::unbounded_channel::<Event>();

        let mut signals = Signals::new(&[SIGHUP, SIGTERM, SIGINT, SIGQUIT])?;
        let signal_handle = signals.handle();
        let signal_sender = tx.clone();
        tokio::spawn(async move {
            // await a signal - once we receive one send the quit event
            let sig = signals.next().await;
            eprintln!("{sig:#?}");
            println!("{sig:#?}");
            info!("{sig:#?}");
            signal_sender.send(Event::Quit).unwrap();
        });

        let tick_rate = std::time::Duration::from_millis(1000);

        let _event_publisher_loop = tokio::spawn(async move {
            let mut crossterm_event_stream = crossterm::event::EventStream::new();
            let mut interval = tokio::time::interval(tick_rate);

            loop {
                let delay = interval.tick();
                // TODO: read about FusedFutures
                let crossterm_event = crossterm_event_stream.next().fuse();

                // TODO: fix unwrap?
                tokio::select! {
                    e = crossterm_event => {
                        match e {
                            Some(Ok(e)) => {
                                tx.send(Event::CrosstermEvent(e)).unwrap();
                            },
                            Some(Err(_e)) => {
                                // tx.send(Event::CrosstermError(Arc::new(e))).unwrap();
                            },
                            None => {},
                        }
                    }
                    _ = delay => {
                        tx.send(Event::Update).unwrap();
                    }
                }
            }
        });

        loop {
            let event = rx.recv().await;

            let mut should_draw = false;

            match event {
                Some(Event::CrosstermEvent(e)) => match e {
                    event::Event::Key(key) => {
                        if key.kind == KeyEventKind::Press
                            && key.code == KeyCode::Char('c')
                            && key.modifiers == KeyModifiers::CONTROL
                        {
                            break;
                        }
                        if key.kind == KeyEventKind::Press && key.code == KeyCode::Char('q') {
                            break;
                        }
                    }
                    _ => {}
                },
                Some(Event::Update) => {}
                Some(Event::Quit) => break,
                Some(Event::Render) => {
                    should_draw = true;
                }
                None => {}
            }

            if should_draw {
                terminal.draw(|frame| {
                    let area = frame.size();
                    frame.render_widget(
                        Paragraph::new("Hello Ratatui! (press 'q' to quit)").white(),
                        area,
                    );
                })?;
            }
        }

        signal_handle.close();

        stdout().execute(LeaveAlternateScreen)?;
        disable_raw_mode()?;

        Ok(())
    }
}
